@@ -10,7 +10,7 @@
 #' @param dates Date. Vector with dates for time features.
 #' @param error_scale String. Scale for the scaled error metrics. Two options: "naive" (average of naive one-step absolute error for the historical series) or "deviation" (standard error of the historical series). Default: "naive".
 #' @param error_benchmark String. Benchmark for the relative error metrics. Two options: "naive" (sequential extension of last value) or "average" (mean value of true sequence). Default: "naive".
-#' @param bounds list of numeric vector, with minimum and maximum bounds for each numeric features. Default: NULL
+#@param bounds list of numeric vector, with minimum and maximum bounds for each numeric features. Default: NULL
 #' @param seed Positive integer. Random seed. Default: 42.
 #'
 #' @author Giancarlo Vercellino \email{giancarlo.vercellino@gmail.com}
@@ -40,12 +40,13 @@
 #' @importFrom modeest mlv1
 #' @import ggplot2
 #' @importFrom moments skewness kurtosis
-#' @importFrom stats quantile sd lm rnorm pnorm fft runif
+#' @importFrom stats quantile sd lm rnorm pnorm fft runif na.omit
 #' @importFrom scales number
 #' @importFrom utils tail head
 #' @import greybox
 #' @importFrom philentropy distance
 #' @importFrom entropy entropy
+#' @import fastDummies
 
 
 #'@examples
@@ -53,7 +54,7 @@
 #'
 #'
 
-spooky <- function(df, seq_len = NULL, lno = NULL, n_samp = 30, n_windows = 3, ci = 0.8, smoother = FALSE, dates = NULL, error_scale = "naive", error_benchmark = "naive", bounds = NULL, seed = 42)
+spooky <- function(df, seq_len = NULL, lno = NULL, n_samp = 30, n_windows = 3, ci = 0.8, smoother = FALSE, dates = NULL, error_scale = "naive", error_benchmark = "naive", seed = 42)
 {
 
   tic.clearlog()
@@ -62,29 +63,23 @@ spooky <- function(df, seq_len = NULL, lno = NULL, n_samp = 30, n_windows = 3, c
   set.seed(seed)
 
   n_length <- nrow(df)
+
+  class_index <- any(map_lgl(df, ~ is.factor(.x) | is.character(.x)))
+  all_classes <- all(class_index)
+  numeric_index <- map_lgl(df, ~ is.integer(.x) | is.numeric(.x))
+  all_numerics <- all(numeric_index)
+  if(!(all_classes | all_numerics)){stop("only all numerics or all classes, not both")}
+
+  if(all_classes){df <- dummy_cols(df, select_columns = NULL, remove_first_dummy = FALSE, remove_most_frequent_dummy = TRUE, ignore_na = FALSE, split = NULL, remove_selected_columns = TRUE); binary_class <- rep(T, ncol(df))}
+  if(all_numerics){binary_class <- rep(F, ncol(df))}
+
+  if(anyNA(df) & all_numerics){df <- as.data.frame(na_kalman(df)); message("kalman imputation on time features\n")}
+  if(anyNA(df) & all_classes){df <- floor(as.data.frame(na_kalman(df))); message("kalman imputation on time features\n")}
+  if(smoother == TRUE & all_numerics){df <- as.data.frame(purrr::map(df, ~ suppressWarnings(loess.as(x=1:n_length, y=.x)$fitted))); message("performing optimal smoothing\n")}
+
   n_feats <- ncol(df)
   feat_names <- colnames(df)
 
-  feat_n_class <- NULL
-  feat_level_names <- NULL
-
-  all_classes <- all(map_lgl(df, ~ is.factor(.x) | is.character(.x)))
-  all_numbers <- all(map_lgl(df, ~ is.numeric(.x) | is.integer(.x)))
-
-  if(!all_classes & !all_numbers){stop("only numbers or factors, but not both")}
-
-  event_class <- all_classes & !all_numbers
-
-  if(event_class == TRUE)
-  {
-    df <- as.data.frame(map(df, ~ as.factor(.x)))
-    feat_level_names <- map(df, ~ levels(.x))
-    feat_n_class <- map_dbl(feat_level_names, ~ length(.x))
-    df <- as.data.frame(data.matrix(df)-1)
-  }
-
-  if(anyNA(df) & event_class == FALSE){df <- as.data.frame(map(df, ~ na_kalman(.x))); message("kalman imputation on target and/or regressors\n")}
-  if(smoother==TRUE & event_class == FALSE){df <- as.data.frame(map(df, ~ suppressWarnings(loess.as(x=1:n_length, y=.x)$fitted))); message("performing optimal smoothing\n")}
 
   if(length(seq_len)==1 & length(lno)==1){n_samp <- 1}
   if(is.null(seq_len)){seq_len <- c(1, round(sqrt(n_length)))}
@@ -92,15 +87,16 @@ spooky <- function(df, seq_len = NULL, lno = NULL, n_samp = 30, n_windows = 3, c
   sqln_set <- round(runif(n_samp, min(seq_len), max(seq_len)))
   lno_set <- round(runif(n_samp, min(lno), max(lno)))
 
-  exploration <- map2(sqln_set, lno_set, ~ windower(df, seq_len = .x, lno = .y, n_windows, ci, dates, error_scale, error_benchmark, feat_n_class, feat_level_names, bounds, seed))
+  exploration <- map2(sqln_set, lno_set, ~ windower(df, seq_len = .x, lno = .y, n_windows, ci, dates, error_scale, error_benchmark, binary_class, seed))
   errors <- round(Reduce(rbind, map(exploration, ~ .x$avg_errors)), 4)
   if(n_samp == 1){errors <- t(as.data.frame(errors))}
   history <- as.data.frame(cbind(seq_len = sqln_set, lno = lno_set, errors))
   rownames(history) <- NULL
 
   if(n_samp > 1){
-    if(event_class == TRUE){history <- history[order(history$czekanowski, history$tanimoto, history$cosine, history$hassebrook, history$jaccard, history$dice, history$canberra, history$gower, history$lorentzian, history$clark), ]}
-    if(event_class == FALSE){history <- history[order(history$me, history$mae, history$mse, history$rmsse, history$mpe, history$mape, history$rmae, history$rame, history$mase, history$smse, history$sce, history$gmrae), ]}
+
+    if(all_numerics == TRUE){history <- ranker(history, focus = -c(1, 2), inverse = NULL, absolute = c("me", "mpe", "sce"), reverse = FALSE)}
+    if(all_classes == TRUE){history <- ranker(history, focus = -c(1, 2), inverse = NULL, absolute = NULL, reverse = FALSE)}
 
     best_index <- as.numeric(rownames(history[1,]))
     best_model <- exploration[[best_index]]
@@ -129,13 +125,9 @@ spooky <- function(df, seq_len = NULL, lno = NULL, n_samp = 30, n_windows = 3, c
   return(outcome)
 }
 
-#' @param feat_n_class Numeric vector. Vector of level numbers for each factor features. Internal.
-#' @param feat_level_names Character vector. Vector collecting the level names for each factor features. Internal.
-#' @param bounds list of numeric vector, with minimum and maximum bounds for each numeric features. Default: NULL
-#' @param seed Positive integer. Random seed. Default: 42.
 
 #########
-windower <- function(df, seq_len, lno = 1, n_windows = 5, ci = 0.8, dates = NULL, error_scale, error_benchmark, feat_n_class = NULL, feat_level_names = NULL, bounds = NULL, seed = 42)
+windower <- function(df, seq_len, lno = 1, n_windows = 5, ci = 0.8, dates = NULL, error_scale, error_benchmark, binary_class, seed)
 {
   n_length <- nrow(df)
   n_feats <- ncol(df)
@@ -143,30 +135,25 @@ windower <- function(df, seq_len, lno = 1, n_windows = 5, ci = 0.8, dates = NULL
   too_short <- floor(n_length/(n_windows + 1)) <= seq_len
   if(too_short){stop("not enough data for the validation windows")}
   idx <- c(rep(1, n_length%%(n_windows + 1)), rep(1:(n_windows + 1), each = n_length/(n_windows + 1)))
-  models <- sapply(1:n_feats, function(f) map(1:n_windows, ~ engine(df[idx <= .x, f, drop = TRUE], seq_len, lno, testing = head(df[idx == .x + 1, f, drop = TRUE], seq_len), ci, dates, feat_names[f], error_scale, error_benchmark, collected_errors = NULL, feat_n_class[f], feat_level_names[[f]], bounds[[f]], seed)), simplify = FALSE)
+  models <- sapply(1:n_feats, function(f) map(1:n_windows, ~ engine(df[idx <= .x, f, drop = TRUE], seq_len, lno, testing = head(df[idx == .x + 1, f, drop = TRUE], seq_len), ci, dates, feat_names[f], error_scale, error_benchmark, collected_errors = NULL, binary_class[f], seed)), simplify = FALSE)
   errors <- map(map_depth(models, 2, ~ .x$errors), ~ round(colMeans(Reduce(rbind, .x)), 4))
   pred_scores <- map(map_depth(models, 2, ~ .x$pred_scores), ~ round(colMeans(Reduce(rbind, .x)), 4))
   collected_errors <- map(map_depth(models, 2, ~ .x$raw_errors), ~ as.data.frame(Reduce(rbind, .x)))
   if(n_feats > 1){avg_errors <- apply(Reduce(rbind, errors), 2, mean)}
   if(n_feats == 1){avg_errors <- unlist(errors)}
 
-  model <- sapply(1:n_feats, function(f) engine(df[, f, drop = TRUE], seq_len, lno, testing = NULL, ci, dates, feat_names[f], error_scale, error_benchmark, collected_errors = collected_errors[[f]], feat_n_class[f], feat_level_names[[f]], bounds[[f]], seed), simplify = FALSE)
+  model <- sapply(1:n_feats, function(f) engine(df[, f, drop = TRUE], seq_len, lno, testing = NULL, ci, dates, feat_names[f], error_scale, error_benchmark, collected_errors = collected_errors[[f]], binary_class[f], seed), simplify = FALSE)
   outcome <- list(errors = errors, avg_errors = avg_errors, pred_scores = pred_scores, model = model)
   return(outcome)
 }
 
 #########
-engine <- function(ts, seq_len, lno = 1, testing = NULL, ci = 0.8, dates = NULL, feat_name, error_scale, error_benchmark, collected_errors, n_class, level_names, bounded, seed)
+engine <- function(ts, seq_len, lno = 1, testing = NULL, ci = 0.8, dates = NULL, feat_name, error_scale, error_benchmark, collected_errors, binary_class, seed)
 {
   orig <- ts
-  is_numeric_sequence <- is.null(n_class) & is.null(level_names)
-  if(!is_numeric_sequence){orig <- level_names[(orig + 1)]}
 
-  if(is_numeric_sequence)
-  {
-    diff_model <- recursive_diff(ts, best_deriv(ts))
-    ts <- diff_model$vector
-  }
+  diff_model <- recursive_diff(ts, best_deriv(ts))
+  ts <- diff_model$vector
 
   n_length <- length(ts)
   transformed <- spectral_transformation(ts)
@@ -175,7 +162,7 @@ engine <- function(ts, seq_len, lno = 1, testing = NULL, ci = 0.8, dates = NULL,
   n_trials <- length(idx_samples)
   cmplx <- transformed$complex
   preds <- Reduce(rbind, map(idx_samples, ~ tail(spectral_inversion(c(cmplx[.x], vector("complex", seq_len))), seq_len)))
-  if(is_numeric_sequence){preds <- t(apply(preds, 1, function(x) tail(invdiff(x, diff_model$tail_value), seq_len)))}
+  preds <- t(apply(preds, 1, function(x) tail(invdiff(x, diff_model$tail_value), seq_len)))
 
   errors <- NULL
   raw_errors <- NULL
@@ -183,23 +170,18 @@ engine <- function(ts, seq_len, lno = 1, testing = NULL, ci = 0.8, dates = NULL,
   plot <- NULL
   quantile_preds <- NULL
   if(!is.null(testing)){
-    errors <- my_metrics(testing, colMeans(preds), ts, error_scale, error_benchmark, n_class)
+    errors <- custom_metrics(testing, colMeans(preds), ts, error_scale, error_benchmark, binary_class)
     raw_errors <- t(replicate(nrow(preds), testing)) - preds
     pred_scores <- prediction_score(preds, testing)
   }
 
   if(!is.null(collected_errors)){
 
-    q_model <- quantile_prediction(raw_pred = NULL, seed_pred = colMeans(preds), raw_error = collected_errors, holdout = NULL, tail_value = NULL, ts, ci, error_scale, error_benchmark, n_class, level_names, bounded, seed)
+    raw_pred <- prediction_integration(seeds = colMeans(preds), raw_errors = collected_errors)
+    q_model <- qpred(raw_pred, holdout_truth = NULL, ts = orig, ci, error_scale, error_benchmark, binary_class, dates, seed)
     quantile_preds <- q_model$quant_pred
-    if(is.null(dates)){hist_dates <- 1:length(orig); forcat_dates <- (length(orig) + 1):(length(orig) + seq_len); rownames(quantile_preds) <- paste0("t", 1:seq_len)}
-    if(!is.null(dates)){hist_dates <- tail(dates, length(orig)); forcat_dates <- seq.Date(tail(dates, 1), tail(dates, 1) + seq_len * mean(diff(dates)), length.out = seq_len); rownames(quantile_preds) <- as.character(forcat_dates)}
-    x_lab <- paste0("Forecasting Horizon for sequence n = ", seq_len)
-    y_lab <- paste0("Forecasting Values for ", feat_name)
-
-    plot <- ts_graph(x_hist = hist_dates, y_hist = orig, x_forcat = forcat_dates, y_forcat = quantile_preds[, 4], lower = quantile_preds[, 2], upper = quantile_preds[, 6],
-                     label_x = x_lab, label_y = y_lab)
 
+    plot <- plotter(quantile_preds, ci, ts = orig, dates, feat_name)
   }
 
   outcome <- list(errors = errors, raw_errors = raw_errors, pred_scores = pred_scores, quantile_preds = quantile_preds, plot = plot)
@@ -248,106 +230,6 @@ best_deriv <- function(ts, max_diff = 3, thresh = 0.001)
 jacknife_index <- function(v, n) {ifelse(n <= 0, n <- 1, n); n <- n - 1; map(1:(length(v) - n), ~ v[- (.x:(.x+n))])}
 
 
-###
-quantile_prediction <- function(raw_pred = NULL, seed_pred = NULL, raw_error = NULL, holdout = NULL, tail_value = NULL, ts, ci,
-                                error_scale = "naive", error_benchmark = "naive", n_class = NULL, level_names = NULL, bounded = NULL, seed = 42)
- {
-  set.seed(seed)
-
-  not_null<- c(!is.null(seed_pred), !is.null(raw_error))
-  if(all(not_null))
-  {
-    pred_integration <- function(pred_seed, error_raw){as.data.frame(map2(pred_seed, error_raw, ~ .x + sample(.y, size = 1000, replace = TRUE)))}
-    raw_pred <- pred_integration(seed_pred, raw_error)
-    if(!is.null(tail_value)){raw_pred <- t(apply(raw_pred, 1, function(x) invdiff(x, heads = tail_value, add = FALSE)))}
-  }
-
-  if(!is.null(raw_pred))
-  {
-    raw_pred <- doxa_filter(ts, raw_pred, n_class, bounded)
-    quants <- sort(unique(c((1-ci)/2, 0.25, 0.5, 0.75, ci+(1-ci)/2)))
-
-    if(is.null(n_class))
-    {
-      p_stats <- function(x){c(min = suppressWarnings(min(x, na.rm = TRUE)), quantile(x, probs = quants, na.rm = TRUE), max = suppressWarnings(max(x, na.rm = TRUE)), mean = mean(x, na.rm = TRUE), sd = sd(x, na.rm = TRUE), mode = suppressWarnings(mlv1(x[is.finite(x)], method = "shorth")), kurtosis = suppressWarnings(kurtosis(x[is.finite(x)], na.rm = TRUE)), skewness = suppressWarnings(skewness(x[is.finite(x)], na.rm = TRUE)))}
-      quant_pred <- as.data.frame(t(as.data.frame(apply(raw_pred, 2, p_stats))))
-      p_value <- apply(raw_pred, 2, function(x) ecdf(x)(seq(min(raw_pred), max(raw_pred), length.out = 1000)))
-      divergence <- c(NA, apply(p_value[,-1, drop = FALSE] - p_value[,-ncol(p_value), drop = FALSE], 2, function(x) abs(max(x, na.rm = TRUE))))
-      upside_prob <- c(NA, colMeans(apply(raw_pred[,-1, drop = FALSE]/raw_pred[,-ncol(raw_pred), drop = FALSE], 2, function(x) x > 1)))
-      iqr_to_range <- (quant_pred[, "75%"] - quant_pred[, "25%"])/(quant_pred[, "max"] - quant_pred[, "min"])
-      median_range_ratio <- (quant_pred[, "max"] - quant_pred[, "50%"])/(quant_pred[, "50%"] - quant_pred[, "min"])
-      quant_pred <- as.data.frame(round(cbind(quant_pred, iqr_to_range, median_range_ratio, upside_prob, divergence), 4))
-    }
-
-    if(is.numeric(n_class) & !is.null(level_names))
-    {
-      freq <- function(x) {sapply(0:(n_class-1), function(n) sum(x == n))}
-      p_stats <- function(x){c(min = suppressWarnings(min(x, na.rm = TRUE)), quantile(x, probs = quants, na.rm = TRUE), max = suppressWarnings(max(x, na.rm = TRUE)), props = freq(x)/length(x), difformity = sd(freq(x)/length(x))/sd(c(1, rep(0, n_class - 1))), entropy = entropy(freq(x)))}
-      quant_pred <- as.data.frame(t(as.data.frame(apply(raw_pred, 2, p_stats))))
-      p_value <- apply(raw_pred, 2, function(x) ecdf(x)(0:(n_class - 1)))
-      divergence <- c(NA, apply(p_value[,-1, drop = FALSE] - p_value[,-ncol(p_value), drop = FALSE], 2, function(x) abs(max(x, na.rm = TRUE))))
-      upgrade_prob <- c(NA, colMeans(apply((raw_pred[,-1, drop = FALSE] + 1)/(raw_pred[,-ncol(raw_pred), drop = FALSE] + 1), 2, function(x) x > 1)))###ADDING ONE TO SOLVE ISSUE WITH CLASS ZERO
-      quant_pred <- as.data.frame(round(cbind(quant_pred, upgrade_prob = upgrade_prob, divergence = divergence), 4))
-
-      ###FIXING FACTOR LABELS
-      n_quants <- length(quants) + 2###QUANTILES + MIN & MAX
-      m_index <- as.matrix(quant_pred[, 1:n_quants] + 1)###INDEXING FROM O TO 1
-      releveled <- as.data.frame(matrix(level_names[m_index], nrow(m_index), ncol(m_index)))
-      colnames(releveled) <- c("min", paste0(quants * 100, "%"), "max")
-      quant_pred <- as.data.frame(cbind(releveled, quant_pred[, c(paste0("props", 1:n_class), "difformity", "entropy", "upgrade_prob", "divergence")]))
-      colnames(quant_pred) <- c(colnames(releveled), paste0("prop_", level_names), "difformity", "entropy", "upgrade_prob", "divergence")
-    }
-  }
-
-  testing_error <- NULL
-  if(!is.null(holdout))
-  {
-    if(is.null(n_class)){
-      mean_pred <- colMeans(raw_pred)
-      testing_error <- my_metrics(holdout, mean_pred, ts, error_scale, error_benchmark, n_class)}
-
-    if(is.numeric(n_class)){
-      testing_error <- apply(apply(raw_pred, 1, function(x) my_metrics(holdout, x, n_class = n_class)), 1, function(x) mean(x, na.rm = T))
-    }
-  }
-
-    outcome <- list(raw_pred = raw_pred, quant_pred = quant_pred, testing_error = testing_error)
-    return(outcome)
-  }
-
-  ###
-  doxa_filter <- function(ts, mat, n_class = NULL, bounded = NULL)
-  {
-    discrete_check <- all(ts%%1 == 0)
-    all_positive_check <- all(ts >= 0)
-    all_negative_check <- all(ts <= 0)
-    monotonic_increase_check <- all(diff(ts) >= 0)
-    monotonic_decrease_check <- all(diff(ts) <= 0)
-
-    monotonic_fixer <- function(x, mode)
-    {
-      model <- recursive_diff(x, 1)
-      vect <- model$vector
-      if(mode == 0){vect[vect < 0] <- 0; vect <- invdiff(vect, model$head_value, add = T)}
-      if(mode == 1){vect[vect > 0] <- 0; vect <- invdiff(vect, model$head_value, add = T)}
-      return(vect)
-    }
-
-    if(is.null(n_class))
-    {
-      if(discrete_check){mat <- round(mat)}
-      if(monotonic_increase_check){mat <- t(apply(mat, 1, function(x) monotonic_fixer(x, mode = 0)))}
-      if(monotonic_decrease_check){mat <- t(apply(mat, 1, function(x) monotonic_fixer(x, mode = 1)))}
-      if(all_positive_check){mat[mat < 0] <- 0}
-      if(all_negative_check){mat[mat > 0] <- 0}
-      if(!is.null(bounded)){mat[mat > max(bounded)] <- max(bounded); mat[mat < min(bounded)] <- min(bounded)}
-    }
-
-    if(is.numeric(n_class)){mat <- round(mat); mat[mat > (n_class - 1)] <- (n_class - 1); mat[mat < 1] <- 0}###REMEMBER TO TRANSFORM CLASS TO DISCRETE STARTING FROM ZERO
-
-    return(mat)
-  }
-
   ###
   recursive_diff <- function(vector, deriv)
   {
@@ -369,48 +251,6 @@ quantile_prediction <- function(raw_pred = NULL, seed_pred = NULL, raw_error = N
     if(add == FALSE){return(vector[-c(1:length(heads))])} else {return(vector)}
   }
 
-  ###
-  my_metrics <- function(holdout, forecast, actuals, error_scale = "naive", error_benchmark = "naive", n_class = NULL)
-  {
-    if(is.null(n_class))
-    {
-      scale <- switch(error_scale, "deviation" = sd(actuals), "naive" = mean(abs(diff(actuals))))
-      benchmark <- switch(error_benchmark, "average" = rep(mean(forecast), length(forecast)), "naive" = rep(tail(actuals, 1), length(forecast)))
-      me <- ME(holdout, forecast, na.rm = TRUE)
-      mae <- MAE(holdout, forecast, na.rm = TRUE)
-      mse <- MSE(holdout, forecast, na.rm = TRUE)
-      rmsse <- RMSSE(holdout, forecast, scale, na.rm = TRUE)
-      mre <- MRE(holdout, forecast, na.rm = TRUE)
-      mpe <- MPE(holdout, forecast, na.rm = TRUE)
-      mape <- MAPE(holdout, forecast, na.rm = TRUE)
-      rmae <- rMAE(holdout, forecast, benchmark, na.rm = TRUE)
-      rrmse <- rRMSE(holdout, forecast, benchmark, na.rm = TRUE)
-      rame <- rAME(holdout, forecast, benchmark, na.rm = TRUE)
-      mase <- MASE(holdout, forecast, scale, na.rm = TRUE)
-      smse <- sMSE(holdout, forecast, scale, na.rm = TRUE)
-      sce <- sCE(holdout, forecast, scale, na.rm = TRUE)
-      gmrae <- GMRAE(holdout, forecast, benchmark, na.rm = TRUE)
-      out <- round(c(me = me, mae = mae, mse = mse, rmsse = rmsse, mpe = mpe, mape = mape, rmae = rmae, rrmse = rrmse, rame = rame, mase = mase, smse = smse, sce = sce, gmrae = gmrae), 3)
-    }
-
-    if(is.numeric(n_class))
-    {
-      czekanowski <- suppressMessages(distance(rbind(holdout, forecast), method = "czekanowski"))
-      tanimoto <- suppressMessages(distance(rbind(holdout, forecast), method = "tanimoto"))
-      cosine <- suppressMessages(1 - distance(rbind(holdout, forecast), method = "cosine"))
-      hassebrook <- suppressMessages(1 - distance(rbind(holdout, forecast), method = "hassebrook"))
-      jaccard <- suppressMessages(distance(rbind(holdout, forecast), method = "jaccard"))
-      dice <- suppressMessages(distance(rbind(holdout, forecast), method = "dice"))
-      canberra <- suppressMessages(distance(rbind(holdout, forecast), method = "canberra"))
-      gower <- suppressMessages(distance(rbind(holdout, forecast), method = "gower"))
-      lorentzian <- suppressMessages(distance(rbind(holdout, forecast), method = "lorentzian"))
-      clark <- suppressMessages(distance(rbind(holdout, forecast), method = "clark"))
-
-      out <- round(c(czekanowski, tanimoto, cosine, hassebrook, jaccard, dice, canberra, gower, lorentzian, clark), 4)
-    }
-
-    return(out)
-  }
 
   ###
   prediction_score <- function(integrated_preds, ground_truth)
@@ -451,3 +291,182 @@ quantile_prediction <- function(raw_pred = NULL, seed_pred = NULL, raw_error = N
 
     return(plot)
   }
+
+  ###
+  qpred <- function(raw_pred, holdout_truth = NULL, ts, ci, error_scale = "naive", error_benchmark = "naive", binary_class = F, dates, seed = 42)
+  {
+    set.seed(seed)
+
+    raw_pred <- doxa_filter(ts, raw_pred, binary_class)
+    quants <- sort(unique(c((1-ci)/2, 0.25, 0.5, 0.75, ci+(1-ci)/2)))
+
+    if(binary_class == F)
+    {
+      p_stats <- function(x){c(min = suppressWarnings(min(x, na.rm = TRUE)), quantile(x, probs = quants, na.rm = TRUE), max = suppressWarnings(max(x, na.rm = TRUE)), mean = mean(x, na.rm = TRUE), sd = sd(x, na.rm = TRUE), mode = suppressWarnings(mlv1(x[is.finite(x)], method = "shorth")), kurtosis = suppressWarnings(kurtosis(x[is.finite(x)], na.rm = TRUE)), skewness = suppressWarnings(skewness(x[is.finite(x)], na.rm = TRUE)))}
+      quant_pred <- as.data.frame(t(as.data.frame(apply(raw_pred, 2, p_stats))))
+      p_value <- apply(raw_pred, 2, function(x) ecdf(x)(seq(min(raw_pred), max(raw_pred), length.out = 1000)))
+      divergence <- c(max(p_value[,1] - seq(0, 1, length.out = 1000)), apply(p_value[,-1, drop = FALSE] - p_value[,-ncol(p_value), drop = FALSE], 2, function(x) abs(max(x, na.rm = TRUE))))
+      upside_prob <- c(mean((raw_pred[,1]/tail(ts, 1)) > 1, na.rm = T), apply(apply(raw_pred[,-1, drop = FALSE]/raw_pred[,-ncol(raw_pred), drop = FALSE], 2, function(x) x > 1), 2, mean, na.rm = T))
+      iqr_to_range <- (quant_pred[, "75%"] - quant_pred[, "25%"])/(quant_pred[, "max"] - quant_pred[, "min"])
+      above_to_below_range <- (quant_pred[, "max"] - quant_pred[, "50%"])/(quant_pred[, "50%"] - quant_pred[, "min"])
+      quant_pred <- round(cbind(quant_pred, iqr_to_range, above_to_below_range, upside_prob, divergence), 4)
+    }
+
+    if(binary_class == T)
+    {
+      p_stats <- function(x){c(min = suppressWarnings(min(x, na.rm = TRUE)), quantile(x, probs = quants, na.rm = TRUE), max = suppressWarnings(max(x, na.rm = TRUE)), prop = mean(x, na.rm = TRUE), sd = sd(x, na.rm = TRUE), entropy = entropy(x))}
+      quant_pred <- as.data.frame(t(as.data.frame(apply(raw_pred, 2, p_stats))))
+      p_value <- apply(raw_pred, 2, function(x) ecdf(x)(c(0, 1)))
+      divergence <- c(max(p_value[,1] - c(0, 1)), apply(p_value[,-1, drop = FALSE] - p_value[,-ncol(p_value), drop = FALSE], 2, function(x) abs(max(x, na.rm = TRUE))))
+      upgrade_prob <- c(mean(((raw_pred[,1] + 1)/tail(ts + 1, 1)) > 1, na.rm = T), apply(apply((raw_pred[,-1, drop = FALSE] + 1)/(raw_pred[,-ncol(raw_pred), drop = FALSE] + 1), 2, function(x) x > 1), 2, mean, na.rm = T))
+      quant_pred <- round(cbind(quant_pred, upgrade_prob = upgrade_prob, divergence = divergence), 4)
+    }
+
+    testing_error <- NULL
+    if(!is.null(holdout_truth))
+    {
+      mean_pred <- colMeans(raw_pred)
+      testing_error <- custom_metrics(holdout_truth, mean_pred, ts, error_scale, error_benchmark, binary_class)
+      pred_scores <- round(prediction_score(raw_pred, holdout_truth), 4)
+      quant_pred <- cbind(quant_pred, pred_scores = pred_scores)
+    }
+
+    if(is.Date(dates))
+    {
+      new_dates<- seq.Date(tail(dates, 1), tail(dates, 1) + nrow(quant_pred) * mean(diff(dates)), length.out = nrow(quant_pred))
+      rownames(quant_pred) <- as.character(new_dates)
+    }
+    else
+    {
+      rownames(quant_pred) <- paste0("t", 1:nrow(quant_pred))
+    }
+
+    outcome <- list(quant_pred = quant_pred, testing_error = testing_error)
+    return(outcome)
+  }
+
+  ###
+  plotter <- function(quant_pred, ci, ts, dates = NULL, feat_name)
+  {
+    seq_len <- nrow(quant_pred)
+    n_ts <- length(ts)
+
+    if(is.Date(dates))
+    {
+      new_dates<- seq.Date(tail(dates, 1), tail(dates, 1) + seq_len * mean(diff(dates)), length.out = seq_len)
+      x_hist <- dates
+      x_forcat <- new_dates
+      rownames(quant_pred) <- as.character(new_dates)
+    }
+    else
+    {
+      x_hist <- 1:n_ts
+      x_forcat <- (n_ts + 1):(n_ts + seq_len)
+      rownames(quant_pred) <- paste0("t", 1:seq_len)
+    }
+
+    quant_pred <- as.data.frame(quant_pred)
+    x_lab <- paste0("Forecasting Horizon for sequence n = ", seq_len)
+    y_lab <- paste0("Forecasting Values for ", feat_name)
+
+    ###if(is.numeric(n_class) & !is.null(level_names)){ts <- level_names[ts + 1]}
+    lower_b <- paste0((1-ci)/2 * 100, "%")
+    upper_b <- paste0((ci+(1-ci)/2) * 100, "%")
+
+    plot <- ts_graph(x_hist = x_hist, y_hist = ts, x_forcat = x_forcat, y_forcat = quant_pred[, "50%"], lower = quant_pred[, lower_b], upper = quant_pred[, upper_b], label_x = x_lab, label_y = y_lab)
+    return(plot)
+  }
+
+  ###
+  doxa_filter <- function(ts, mat, binary_class = F)
+  {
+    discrete_check <- all(ts%%1 == 0)
+    all_positive_check <- all(ts >= 0)
+    all_negative_check <- all(ts <= 0)
+    monotonic_increase_check <- all(diff(ts) >= 0)
+    monotonic_decrease_check <- all(diff(ts) <= 0)
+
+    monotonic_fixer <- function(x, mode)
+    {
+      model <- recursive_diff(x, 1)
+      vect <- model$vector
+      if(mode == 0){vect[vect < 0] <- 0; vect <- invdiff(vect, model$head_value, add = TRUE)}
+      if(mode == 1){vect[vect > 0] <- 0; vect <- invdiff(vect, model$head_value, add = TRUE)}
+      return(vect)
+    }
+
+    if(all_positive_check){mat[mat < 0] <- 0}
+    if(all_negative_check){mat[mat > 0] <- 0}
+    if(discrete_check){mat <- round(mat)}
+    if(monotonic_increase_check){mat <- t(apply(mat, 1, function(x) monotonic_fixer(x, mode = 0)))}
+    if(monotonic_decrease_check){mat <- t(apply(mat, 1, function(x) monotonic_fixer(x, mode = 1)))}
+
+    if(binary_class == T){mat[mat > 1] <- 1; mat[mat < 1] <- 0}
+    mat <- na.omit(mat)
+
+    return(mat)
+  }
+
+
+  ###
+  custom_metrics <- function(holdout, forecast, actuals, error_scale = "naive", error_benchmark = "naive", binary_class = F)
+  {
+
+    if(binary_class == F)
+    {
+      scale <- switch(error_scale, "deviation" = sd(actuals), "naive" = mean(abs(diff(actuals))))
+      benchmark <- switch(error_benchmark, "average" = rep(mean(forecast), length(forecast)), "naive" = rep(tail(actuals, 1), length(forecast)))
+      me <- ME(holdout, forecast, na.rm = TRUE)
+      mae <- MAE(holdout, forecast, na.rm = TRUE)
+      mse <- MSE(holdout, forecast, na.rm = TRUE)
+      rmsse <- RMSSE(holdout, forecast, scale, na.rm = TRUE)
+      mre <- MRE(holdout, forecast, na.rm = TRUE)
+      mpe <- MPE(holdout, forecast, na.rm = TRUE)
+      mape <- MAPE(holdout, forecast, na.rm = TRUE)
+      rmae <- rMAE(holdout, forecast, benchmark, na.rm = TRUE)
+      rrmse <- rRMSE(holdout, forecast, benchmark, na.rm = TRUE)
+      rame <- rAME(holdout, forecast, benchmark, na.rm = TRUE)
+      mase <- MASE(holdout, forecast, scale, na.rm = TRUE)
+      smse <- sMSE(holdout, forecast, scale, na.rm = TRUE)
+      sce <- sCE(holdout, forecast, scale, na.rm = TRUE)
+      gmrae <- GMRAE(holdout, forecast, benchmark, na.rm = TRUE)
+      out <- round(c(me = me, mae = mae, mse = mse, rmsse = rmsse, mpe = mpe, mape = mape, rmae = rmae, rrmse = rrmse, rame = rame, mase = mase, smse = smse, sce = sce, gmrae = gmrae), 3)
+    }
+
+    if(binary_class == T)
+    {
+      dice <- suppressMessages(distance(rbind(holdout, forecast), method = "dice"))
+      jaccard <- suppressMessages(distance(rbind(holdout, forecast), method = "jaccard"))
+      cosine <- suppressMessages(distance(rbind(holdout, forecast), method = "cosine"))
+      canberra <- suppressMessages(distance(rbind(holdout, forecast), method = "canberra"))
+      gower <- suppressMessages(distance(rbind(holdout, forecast), method = "gower"))
+      tanimoto <- suppressMessages(distance(rbind(holdout, forecast), method = "tanimoto"))
+      hassebrook <- 1 - suppressMessages(distance(rbind(holdout, forecast), method = "hassebrook"))
+      taneja <- suppressMessages(distance(rbind(holdout, forecast), method = "taneja"))
+      lorentzian <- suppressMessages(distance(rbind(holdout, forecast), method = "lorentzian"))
+      clark <- suppressMessages(distance(rbind(holdout, forecast), method = "clark"))
+      sorensen <- suppressMessages(distance(rbind(holdout, forecast), method = "sorensen"))
+      harmonic_mean <- suppressMessages(distance(rbind(holdout, forecast), method = "harmonic_mean"))
+      avg <- suppressMessages(distance(rbind(holdout, forecast), method = "avg"))
+
+      out <- round(c(dice, jaccard, cosine, canberra, gower, tanimoto, hassebrook, taneja, lorentzian, clark, sorensen, harmonic_mean, avg), 4)
+    }
+
+    return(out)
+  }
+
+  ###
+  prediction_integration <- function(seeds, raw_errors){as.matrix(as.data.frame(map2(seeds, raw_errors, ~ .x + sample(.y, size = 1000, replace = TRUE))))}
+
+  ###
+  ranker <- function(df, focus, inverse = NULL, absolute = NULL, reverse = FALSE)
+  {
+    rank_set <- df[, focus, drop = FALSE]
+    if(!is.null(inverse)){rank_set[, inverse] <- - rank_set[, inverse]}###INVERSION BY COL NAMES
+    if(!is.null(absolute)){rank_set[, absolute] <- abs(rank_set[, absolute])}###ABS BY COL NAMES
+    index <- apply(scale(rank_set), 1, mean, na.rm = TRUE)
+    if(reverse == FALSE){df <- df[order(index),]}
+    if(reverse == TRUE){df <- df[order(-index),]}
+    return(df)
+  }
+
